
// Schedules for irregular subscripting of distributed arrays.

#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include "MPIinclude.h"
#include "globals.h"
#include "threads.h"
#include "ProcTree.h"
#include "Procs.h"
#include "Range.h"
#include "DAD.h"
#include "schedule.h"
#include "MessSchedule.h"
#include "TreeSchedule.h"
#include "DataSchedule.h"
#include "Reduce.h"
#include "Gather.h"

//#include "admacros.h"


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `GatherSkeleton'                                              //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

GatherSkeleton :: GatherSkeleton(const DAD* _loc, const DAD* _rem,
                                 const int len,
                                 const DAD** _subs, int** subsData,
                                 const DAD* _msk, int* mskData) :
      DataSchedule(len),
      loc(_loc), rem(_rem), subs(_subs), msk(_msk), rank(_loc->rnk()) {

  Group p = loc->grp() ;
  restrictToAPG(p) ;
  q = rem->grp() ;
  restrictToAPG(q) ;

  if(p.member()) {
    DimensionSet rep = p.dims() - loc->sig() ;  // Spans replication group.

    Group reduced = p ;
    reduced.restrict(rep) ;  // Eliminate dims in which data is replicated.

    if(reduced.lead_prc() == p.lead_prc())  // Am part of lead reduced group.
      loop(0, subsData, mskData, 0) ;
  }

  TreeSchedule spanTree ;
  //apgSpanTree(&spanTree) ;
  :: spanTree(&spanTree, apg) ;
  DataSchedule :: build(&spanTree) ;
}

void GatherSkeleton :: loop(int locOffset, int** subsElem, int* mskElem,
                            const int r) {

  if(r == rank) {
    if(!msk || *mskElem) {
      Group  remGrp = q ;
      int remOffset = 0 ;

      for(int s = 0 ; s < rem->rnk() ; s++) {
	Range y(rem->rng(s)) ;
	Location i(y, *(subsElem [s])) ;

        remOffset += rem->str(s).offset(i) ;
        remGrp    /= i ;
      }

      transReq(locOffset, remGrp.lead_abs(), remOffset) ;
    }
  }
  else {
    LocBlocksIndex i(loc->rng(r)) ;

    Stride locStr = loc->str(r) ;
    Stride mskStr ;
    if(msk) mskStr = msk->str(r) ;

    int** subsDataElem = new int* [rem->rnk()] ;
    int* subsStep = new int [rem->rnk()] ;

    for(i.beginLocBlk() ; i.test() ; i.nextLocBlk()) {

      int s ;

      int locOffsetElem = locOffset + locStr.offset(i) ;
      const int locStep = locStr.step(i) ;

      for(s = 0 ; s < rem->rnk() ; s++) {
        Stride subStr = (subs [s])->str(r) ;

        subsDataElem [s] = subsElem [s] + subStr.offset(i) ;
        subsStep [s]     = subStr.step(i) ;
      }

      int* mskDataElem  = msk ? mskElem + mskStr.offset(i) : (int*) NULL ;
      const int mskStep = msk ? mskStr.step(i) : 0 ;

      for(int i_block = 0 ; i_block < i.count ; i_block++) {
        loop(locOffsetElem, subsDataElem, mskDataElem, r + 1) ;

	locOffsetElem += locStep ;
        for(s = 0 ; s < rem->rnk() ; s++)
          subsDataElem [s] += subsStep [s] ;
        if(msk)
          mskDataElem += mskStep ;
      }
    }

    delete [] subsStep ;
    delete [] subsDataElem ;
  }
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `ReshapeSkeleton'                                             //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

ReshapeSkeleton :: ReshapeSkeleton(const DAD* _loc, const DAD* _rem,
                                   const int len) :
		       loc(_loc), rem(_rem), rank(_loc->rnk()),
		       DataSchedule(len) {

  Group p = loc->grp() ;
  q = rem->grp() ;
  restrictToAPG(q) ;

  if(p.member()) {
    DimensionSet rep = p.dims() - loc->sig() ;  // Spans replication group.

    Group reduced = p ;
    reduced.restrict(rep) ;  // Eliminate dims in which data is replicated.

    if(reduced.lead_prc() == p.lead_prc())  // Am part of lead reduced group.
      loop(0, 0, 1, 0) ;
  }

  TreeSchedule spanTree ;
  //apgSpanTree(&spanTree) ;
  :: spanTree(&spanTree, apg) ;
  DataSchedule :: build(&spanTree) ;
}


void ReshapeSkeleton :: loop(int locOffset, int seqIndex, int seqStride,
                             int r) {

  if(r == rank) {
    Group remGrp = q ;
    int remOffset = 0 ;

    for(int s = 0 ; s < rem->rnk() ; s++) {
      Range y(rem->rng(s)) ;
      Location i(y, seqIndex % y.size()) ;

      remOffset += rem->str(s).offset(i) ;
      remGrp    /= i ;
    }

    transReq(locOffset, remGrp.lead_abs(), remOffset) ;
  }
  else {
    Range x = loc->rng(r) ;
    Index i(x) ;

    for(i.begin() ; i.test() ; i.next()) {
      int locOffsetElem = locOffset + loc->str(r).offset(i) ;

      loop(locOffsetElem,
           seqIndex + seqStride * i, x.size() * seqStride,
	   r + 1) ;
    }
  }
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `ReshapePadSkeleton'                                          //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

ReshapePadSkeleton :: ReshapePadSkeleton(const DAD* _loc, int _remSize,
                                         const DAD* _pad,
                                         const int len) :
			  loc(_loc), remSize(_remSize), pad(_pad),
                          rank(_loc->rnk()),
			  DataSchedule(len) {

  // Silly `reshape' padding to support Fortran.

  Group p = loc->grp() ;
  q = pad->grp() ;
  restrictToAPG(q) ;

  if(p.member()) {
    DimensionSet rep = p.dims() - loc->sig() ;  // Spans replication group.

    Group reduced = p ;
    reduced.restrict(rep) ;  // Eliminate dims in which data is replicated.

    if(reduced.lead_prc() == p.lead_prc())  // Am part of lead reduced group.
      loop(0, 0, 1, 0) ;
  }

  TreeSchedule spanTree ;
  //apgSpanTree(&spanTree) ;
  :: spanTree(&spanTree, apg) ;
  DataSchedule :: build(&spanTree) ;
}

void ReshapePadSkeleton :: loop(int locOffset, int seqIndex, int seqStride,
                                int r) {

  if(r == rank) {
    if(seqIndex >= remSize) {
      Range y(pad->rng(0)) ;
      const int padIndex = (seqIndex - remSize) % y.size() ;
 
      Location i(y, padIndex) ;

      Group remGrp = q / i ;
      transReq(locOffset, remGrp.lead_abs(), pad->str(0).offset(i)) ;
    }
  }
  else {
    Range x = loc->rng(r) ;
    Index i(x) ;

    for(i.begin() ; i.test() ; i.next()) {
      loop(locOffset + loc->str(r).offset(i),
	   seqIndex + seqStride * i, x.size() * seqStride,
	   r + 1) ;
    }
  }
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `VecGatherSkeleton'                                           //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


int VecGatherSkeleton :: blockRank(const int rank, const DAD** subs) {
  int result = 0 ;
  for (int s = 0 ; s < rank ; s++)
    if(!(subs [s])) result++ ;

  return result ;
}

VecGatherSkeleton :: VecGatherSkeleton(const DAD* _loc, const DAD* _rem,
                                       const int len,
                                       const DAD** _subs, int** _subsData) :
                         BlockDataSchedule(len,
                                           blockRank(_rem->rnk(), _subs)),
                         loc(_loc), rem(_rem),
                         subs(_subs), subsData(_subsData),
                         rank(_loc->rnk()) {
  Group p = loc->grp() ;
  restrictToAPG(p) ;
  Group q = rem->grp() ;
  restrictToAPG(q) ;

  if(p.member()) {
    DimensionSet rep = p.dims() - loc->sig() ;  // Spans replication group.

    Group reduced = p ;
    reduced.restrict(rep) ;  // Eliminate dims in which data is replicated.

    if(reduced.lead_prc() == p.lead_prc()) {  // Am part of lead reduced group.
      exts     = new int [BlockDataSchedule :: rank] ;
      locSteps = new int [BlockDataSchedule :: rank] ;
      remSteps = new int [BlockDataSchedule :: rank] ;

      loop(0, 1, 0, 1, 1, q, 0, 0) ;

      delete [] remSteps ;
      delete [] locSteps ;
      delete [] exts ;
    }
  }

  TreeSchedule spanTree ;
  //apgSpanTree(&spanTree) ;
  :: spanTree(&spanTree, apg) ;
  BlockDataSchedule :: build(&spanTree) ;
}

void VecGatherSkeleton :: loop(const int locOffset, const int locCtg,
                               const int remOffset, const int remCtg,
                               const int vol,
                               Group remGrp,
                               const int r, const int s) {

  if(r == rank)
    transReq(locOffset, locSteps, locCtg, remOffset, remSteps, remCtg,
             exts, vol, remGrp.lead_abs()) ;
  else {

    Range  x      = loc->rng(r) ;
    Stride locStr = loc->str(r) ;

    Range  y      = rem->rng(r) ;
    Stride remStr = rem->str(r) ;

    const DAD* sub = subs [r] ;

    if(sub) {

      // Dimension has a vector subscript.

      Index i(x) ;

      for(i.begin() ; i.test() ; i.next()) {
        int* subData = subsData [r] ;
        Location j(y, subData [sub->str(0).offset(i)]) ;

	loop(locOffset + locStr.offset(i), locCtg,
             remOffset + remStr.offset(j), remCtg,
             vol,
             remGrp / j, r + 1, s) ;
      }
    }
    else {

      // Dimension is unsubscripted.  (cf `remap').

      LocBlocksIndex i(x) ;

      for(i.beginLocBlk() ; i.test() ; i.nextLocBlk()) {

	int locOffsetElem = locOffset + locStr.offset(i) ;
	const int locStep = locStr.step(i) ;

        AllBlocksIndex j(y.subrng(i.count, i.glb_bas, i.glb_stp)) ;

	for(j.beginAllBlk() ; j.test() ; j.nextAllBlk()) {

	  const int remOffsetElem = remOffset + remStr.offset(j) ;
	  const int remStep       = remStr.step(j) ;

          const int jlocStep = locStep * j.glb_stp ;

	  exts     [s] = j.count ;
          locSteps [s] = jlocStep ;
	  remSteps [s] = remStep ;

	  loop(locOffsetElem + j.glb_bas * locStep, locCtg && jlocStep == vol,
               remOffsetElem, remCtg && remStep == vol,
               vol * j.count,
               remGrp / j,
               r + 1, s + 1) ;
	}
      }
    }
  }
}


// -------------------------------------------------------------
// Things to do
//
//   Unify `Reshape' and `ReshapePad' by introducing a start offset in former?
//
//   Could `reshape' be generalized to support arbitrary F77 ``sequence
//       association'', somehow?
//
//   In any case, `reshape' should be based on a `MessSchedule'...
//       it is ``deterministic''.
//
//   Add a version of `VecGatherSkeleton' that works on a `DataSchedule',
//       requiring all dimensions to be vector subscripted.
//
//   Eliminate use of `Index' iterators in in `VecGatherSkeleton' and
//       `Reshape' (in favour of lower-level `LocBlocksIndex').
//
//   Where there are several short `int' vectors in a schedule (`exts',
//     `steps', ...) allocate and deallocate them as a single vector.
//
//     Or allocate them using `Vec'?


